"
From: carl@ParcPlace.com (Carl Watts)
Newsgroups: comp.lang.smalltalk
Subject: Re: Fastest collection for use other than lookup?
Followup-To: comp.lang.smalltalk
Date: Tue, 28 Jun 1994 14:12:34 -0800
Organization: ParcPlace Systems, Inc.
Lines: 445
Message-ID: <carl-280694141234@opus.parcplace.com>
References: <Crx6tw.Mvz@freenet.carleton.ca> <2uk7ls$72c@panix.com>
NNTP-Posting-Host: opus.parcplace.com

It isn't easy to get much faster than OrderedCollection for just adding to
a collection (using add:) and then enumerating over the collection (using
do:).

Following is an implementation of a new collection class called Chain that
is like OrderedCollection except it doesn't have to grow by copying like
OrderedCollection does.  Chain is about as fast as you can get at adding
items to a collection and enumerating over them.  The price you pay for
this is space.  Chain is far less space efficient than OrderedCollection.

There are some benchmark methods on Chain class that will allow you to
compare the performance of Chain and OrderedCollection.  You will see that
Chain isn't significantly faster than OrderedCollection for adding and
enumerating though it can be faster for other operations.

The moral is:  Know what the classes you use are doing for you, and what
they aren't.

claus: also needs ChainLink to be filed-in
       In ST/X, OrderedCollection is the better choice most of the
       time (OC is pretty fast - indeed). The benchmarks here all run between
       2 and 5 times faster with OC's compared to Chains ....
"
!

Collection subclass: #Chain
	instanceVariableNames: 'lastLink firstLink '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Linked'!

Chain comment:
'Chains are a collection of arbitrary elements explicitly ordered by the
sequence in which elements are added and removed or inserted.  Each element
is stored using a ChainLink.  These links can be traversed efficiently
forwards or backwards using nextLink and previousLink.  The ChainLinks are
linked as a ring but an individual ChainLink can efficiently determine if
it is the first or last link in the Chain.  So the Chain can act as a ring
or as non-ring.

Chain provides many of the same abilities as OrderedCollection except Chain
is not indexable.  OrderedCollections are much more space efficient but
Chains are faster for some addition/removal operations.

Adding a new link before or after any given link or at the beginning or end
of the Chain is an Order(1) operation.  Removing any given link is an
Order(1) operation.  For OrderedCollections all these operations (except
removeLast) are Order(n).

This class and its private link class ChainLink were implemented by Carl
Watts (Carl@ParcPlace.com).

Instance Variables:
	firstLink  <ChainLink or self>  The first link in the chain or self if
empty.
	lastLink  <ChainLink or self>  The last link in the chain or self if
empty.'!


!Chain methodsFor: 'private'!

linkAfterPreviousLink
	"Answer the link after the receiver's previous link."
	"Double-Dispatched method from ChainLink.  For the receiver answer the
first link."

	^firstLink!

linkBeforeNextLink
	"Answer the link before the receiver's next link."
	"Double-Dispatched method from ChainLink.  For the receiver answer the
last link."

	^lastLink!

nextLinkIsFirst
	"Answer if the nextLink is the first link in the chain."
	"Double-dispatched from a ChainLink.  I know I can answer true."

	^true!

previousLink: aLink
	"Set the ChainLink before the receiver."
	"Double-dispatched method from a ChainLink.  Set aLink as the first link."

	firstLink := aLink!

previousLinkIsLast
	"Answer if the previousLink is the last link in the chain."
	"Double-dispatched from a ChainLink.  I know I can answer true."

	^true!

setNextLink: aLink
	"Set the ChainLink before the receiver."
	"Double-dispatched method from a ChainLink.  Set aLink as the first link."

	firstLink := aLink!

setPreviousLink: aLink
	"Set the ChainLink before the receiver."
	"Double-dispatched method from a ChainLink.  Set aLink as the last link."

	lastLink := aLink! !

!Chain methodsFor: 'accessing'!

first
	"Answer the first value;  provide an error notification if the receiver is
empty."

	^self firstLink value!

firstLink
	"Answer the first link;  provide an error notification if the receiver is
empty."

	self emptyCheck.
	^firstLink!

last
	"Answer the last value;  provide an error notification if the receiver is
empty."

	^self lastLink value!

lastLink
	"Answer the last link;  provide an error notification if the receiver is
empty."

	self emptyCheck.
	^lastLink!

linkAtValue: aValue
	"Answer a link with value = aValue."

	| aLink |
	self isEmpty ifTrue: [^self].
	aLink := firstLink.
	[aLink value = aValue ifTrue: [^aLink].
	 aLink := aLink nextLink.
	 aLink == firstLink] whileFalse.
	self notFoundError!

size
	"Answer how many elements the receiver contains."

	| tally |
	tally := 0.
	self do: [:each | tally := tally + 1].
	^tally! !

!Chain methodsFor: 'adding'!

add: aValue
	"Add aValue to the end of the receiver.  Answer aValue."

	^self addLast: aValue!

addFirst: aValue
	"Add aValue to the beginning of the receiver. Answer aValue"

	self addLinkFirst: (ChainLink value: aValue).
	^aValue!

addLast: aValue
	"Add aValue to the end of the receiver. Answer aValue."

	self addLinkLast: (ChainLink value: aValue).
	^aValue!

addLink: aLink
	"Add aLink to the end of the receiver.  Answer aLink."

	^self addLinkLast: aLink!

addLinkFirst: aLink
	"Add aLink to the beginning of the receiver."

	aLink previousLink: self nextLink: firstLink.
	^aLink!

addLinkLast: aLink
	"Add aLink to the end of the receiver."

	aLink previousLink: lastLink nextLink: self.
	^aLink! !

!Chain methodsFor: 'initialization'!

initialize
	"Initialize the receiver."

	firstLink := lastLink := self! !

!Chain methodsFor: 'enumerating'!

do: aBlock
	"Evaluate aBlock with each of the receiver's values as the argument."

	| aLink |
	self isEmpty ifTrue: [^self].
	aLink := firstLink.
	[aBlock value: aLink value.
	 aLink := aLink nextLink.
	 aLink == firstLink] whileFalse!

linksDo: aBlock
	"Evaluate aBlock with each of the receiver's links as the argument."

	| aLink |
	self isEmpty ifTrue: [^self].
	aLink := firstLink.
	[aBlock value: aLink.
	 aLink := aLink nextLink.
	 aLink == firstLink] whileFalse! !

!Chain methodsFor: 'testing'!

isEmpty
	"Answer whether the receiver contains any elements."

	^firstLink == self! !

!Chain methodsFor: 'removing'!

remove: oldObject ifAbsent: anExceptionBlock 
	"Remove oldObject as one of the receiver's elements.  Answer oldObject
unless no element is equal to oldObject, in which case, evaluate
anExceptionBlock."

	(self linkAtValue: oldObject) remove.
	^oldObject! !

Chain class
	instanceVariableNames: ''!


!Chain class methodsFor: 'instance creation'!

new
	"Answer with a new instance of the receiver."

	^super new initialize! !

!Chain class methodsFor: 'examples'!

exampleBenchmark1
	"This benchmark compares an operation that both Chain and
OrderedCollection are good at, adding at the end.  Their times should be
comparable."

	self runBenchMark: [:collectionClass |
		| aCollection |
		20 timesRepeat: [
			aCollection := collectionClass new.
			aCollection addAll: (1 to: 1000)]]

	"
	 Chain exampleBenchmark1
	"
!

exampleBenchmark2
	"This benchmark compares operations that both Chain and OrderedCollection
are good at, adding at the end and enumerating.  Their times should be
comparable."

	self runBenchMark: [:collectionClass |
		| aCollection |
		20 timesRepeat: [
			aCollection := collectionClass new.
			aCollection addAll: (1 to: 1000).
			aCollection do: [:each | each + 1]]]
	"
	 Chain exampleBenchmark2
	"
!

exampleBenchmark3
	"This benchmark shows that Chain is significantly faster when removing
elements from the middle even though both Chain and OrderedCollection have
to do a linear search to find the element."

	self runBenchMark: [:collectionClass |
		| aCollection |
		8 timesRepeat: [
			aCollection := collectionClass new.
			aCollection addAll: (1 to: 1000).
			aCollection removeAll: (250 to: 750)]]
	"
	 Chain exampleBenchmark3
	"
!

runBenchMark: benchMarkBlock
	"Run a benchmark with the given test block.  The block will be evaluated
with different collection classes as the parameter.  The time required to
evaluate the block will be compared."

	| chainTime orderedCollectionTime |
	chainTime := Time millisecondsToRun: [benchMarkBlock value: Chain].
	Transcript cr; cr; nextPutAll: 'Chain time '; print: chainTime; endEntry.
	orderedCollectionTime := Time millisecondsToRun: [benchMarkBlock value:
OrderedCollection].
	Transcript cr; nextPutAll: 'OrderedCollection time '; print:
orderedCollectionTime; endEntry.
	Transcript cr; nextPutAll: 'Chain is '; print: orderedCollectionTime /
chainTime asFloat; nextPutAll: ' times faster for this benchmark';
endEntry! !
