"
From: carl@ParcPlace.com (Carl Watts)
Newsgroups: comp.lang.smalltalk
Subject: Re: Fastest collection for use other than lookup?
Followup-To: comp.lang.smalltalk
Date: Tue, 28 Jun 1994 14:12:34 -0800
Organization: ParcPlace Systems, Inc.
Lines: 445
Message-ID: <carl-280694141234@opus.parcplace.com>
References: <Crx6tw.Mvz@freenet.carleton.ca> <2uk7ls$72c@panix.com>
NNTP-Posting-Host: opus.parcplace.com

It isn't easy to get much faster than OrderedCollection for just adding to
a collection (using add:) and then enumerating over the collection (using
do:).

Following is an implementation of a new collection class called Chain that
is like OrderedCollection except it doesn't have to grow by copying like
OrderedCollection does.  Chain is about as fast as you can get at adding
items to a collection and enumerating over them.  The price you pay for
this is space.  Chain is far less space efficient than OrderedCollection.

There are some benchmark methods on Chain class that will allow you to
compare the performance of Chain and OrderedCollection.  You will see that
Chain isn't significantly faster than OrderedCollection for adding and
enumerating though it can be faster for other operations.

The moral is:  Know what the classes you use are doing for you, and what
they aren't.

claus: this class is needed for Chain
"

Object subclass: #ChainLink
        instanceVariableNames: 'value nextLink previousLink '
        classVariableNames: ''
        poolDictionaries: ''
        category: 'Collections-Support'!
ChainLink comment:
'ChainLinks are individual elements in a doubly-linked chain or ring. 
Chains use ChainLinks to hold their elements.  Each ChainLink has an
arbitrary object as its value.

Instance Variables:
        nextLink                <ChainLink or Chain>  The ChainLink after this one or the Chain
itself if this is the last link.
        previousLink    <ChainLink or Chain>  The ChainLink before this one or the
Chain itself if this is the first link.
        value                   <Object>  The arbitrary object associated with this link.'!


!ChainLink methodsFor: 'accessing'!

nextLink
        "Answer the next ChainLink after the receiver."

        ^nextLink linkAfterPreviousLink!

previousLink
        "Answer the ChainLink before the receiver."

        ^previousLink linkBeforeNextLink!

value
        "Answer the value associated with the receiver."

        ^value!

value: anObject
        "Set the value associated with the receiver."

        value := anObject! !

!ChainLink methodsFor: 'private'!

linkAfterPreviousLink
        "Answer the link after the receiver's previous link."
        "Double-Dispatched Method.  For the receiver answer self."

        ^self!

linkBeforeNextLink
        "Answer the link before the receiver's next link."
        "Double-Dispatched Method.  For the receiver answer self."

        ^self!

nextLinkIsFirst
        "Answer if the nextLink is the first link in the chain."
        "Double-dispatched from a ChainLink. The next link can't be the last link
in the chain because I'm before it."

        ^false!

previousLink: beforeLink nextLink: afterLink
        "Set the links before and after the receiver."

        previousLink := beforeLink.
        nextLink := afterLink.
        previousLink setNextLink: self.
        nextLink setPreviousLink: self!

previousLinkIsLast
        "Answer if the previousLink is the last link in the chain."
        "Double-dispatched from another ChainLink.  The previous link can't be the
last link in the chain because I'm after it."

        ^false!

setNextLink: aLink
        "Set the next ChainLink after the receiver."

        nextLink := aLink!

setPreviousLink: aLink
        "Set the ChainLink before the receiver."

        previousLink := aLink! !

!ChainLink methodsFor: 'removing'!

remove
        "Remove the receiver from the Chain."

        previousLink setNextLink: nextLink.
        nextLink setPreviousLink: previousLink.
        nextLink := previousLink := nil! !

!ChainLink methodsFor: 'inserting'!

afterInsert: aValue
        "Insert aValue as the value after the receiver in the same chain."

        self afterInsertLink: (ChainLink value: aValue)!

afterInsertLink: aLink
        "Insert aLink as the link after the receiver."

        aLink previousLink: self nextLink: nextLink!

beforeInsert: aValue
        "Insert aValue as the link before the receiver in the same Chain."

        self beforeInsertLink: (ChainLink value: aValue)!

beforeInsertLink: aLink
        "Insert aLink as the link before the receiver in the same Chain."

        aLink previousLink: previousLink nextLink: self! !

!ChainLink methodsFor: 'testing'!

isFirst
        "Answer if the receiver is the first link in its chain."

        ^previousLink nextLinkIsFirst!

isLast
        "Answer if the receiver is the last link in its chain."

        ^nextLink previousLinkIsLast! !

ChainLink class
        instanceVariableNames: ''!


!ChainLink class methodsFor: 'instance creation'!

value: aValue
        "Answer a new instance with a value of aValue."

        ^self basicNew value: aValue! !
