'Copyright (c) 1991 by Bill Voss.  All rights reserved.'!

"
(Smalltalk includesKey: #UnixSocketAccessor) ifFalse:[
    (Smalltalk includesKey: #Socket) ifFalse:[
        self error: 'utils/UnixSock.st must be filed in first'
    ]
]
"
!

Object subclass: #RDoItServer
        instanceVariableNames: 'socket myPortNumber connectProcess'
        classVariableNames: 'BadHosts OkHosts '
        poolDictionaries: ''
        category: 'Voss-RDoIt'
!

!RDoItServer class methodsFor:'documentation'!

documentation
"
    'First how to use:
        [RDoItServer initialize.]   To reset the authorized/unauthorized host sets.
        [RDoItServer start.]    or    
            [RDoItServer startAt: 8004.]    To start a server on port (default 8004).
        [RDoItServer killAll.]              To kill all servers.

    ----------------------------------------------------------

    This is a quick hack to allow an external C program to execute a Smalltalk doIt.

       I built this so I could put the Launcher options on my X11 root menu.
       I also wanted a tiny project to help me get back up to speed after
       not writing any Smalltalk code for a long time.

    The basic idea is that we start up a Smalltalk process which creates a socket,
    and sits there listening for connections.  When a connection comes in, it forks
    a subprocess to do the actual work.  This subprocess reads from the socket
    until it sees a CR, and then evaluates what it has read (simillar to DoIt in a workspace).
    The subprocess then closes its connection and dies, while the main process
    continues to wait for connections.   
    The server also catches the #returnFromSnapshot update: message, and restarts itself.

    The external C program 'rdoit hostname portnumber message' is similar to 'rsh' in operation.

    CAUTION: The doits are NOT logged in the change log.
    DANGER: Security is done on a host by host basis.  When a connection
            comes in from a previously unrecognized host, a dialog box
            pops up asking the user to authorize the host.
                Select yes, and all future requests from that host will suceed.
                Select no, and all future attempts will be denied.
                (This is done with the Sets BadHosts and OkHosts.)
            This means that while the server is running, ANY user on any
            authorized host could have your st80 process do anything.
            Including run any program on the system SO BE CAREFUL !!

    FUTURE WORK:  
            Fix security problems.
                (There are only two people with accounts on my system,
                 so I personally am not going to 'fix' this problem anytime soon.)
            Cleanup Code.
                (Especially consider logging the doits.)
            Find and Fix the Bugs.
                (I haven''t used the UnixSockets fileIn before, so I almost
                 certainly goofed somewhere.)

    Bill Voss     <voss@cs.uiuc.edu>     June 14th, 1991
"
! !

!RDoItServer methodsFor: 'initialize-release'!

release
"claus:ORIGINAL
        (socket isKindOf: UnixSocketAccessor) ifTrue: [socket close].
"
        socket notNil ifTrue: [socket close].
        socket := nil.
        super release.
        connectProcess notNil ifTrue:[
            connectProcess terminate.
            connectProcess := nil
        ]
! !

!RDoItServer methodsFor: 'restart-update'!

restart
        "Called when snapshot is started, and at instance creation."

"claus:original
        [self loopForever] forkAt: self serverPriority.
"
        connectProcess := [self loopForever] forkAt: self serverPriority.
        connectProcess name:'rdoit server'.

!

startServer
        "This is only run by SERVER instances at startup."

        Smalltalk addDependent: self.
        self restart.
        ^self
!

startServerAt: aPortNumber 
        "This is only run by SERVER instances at startup."

        myPortNumber := aPortNumber.
        ^ self startServer
!

update: aSymbol 
        "Dependents of SystemDictionary Smalltalk are sent update: 
        #returnFromSnapshot when a snapshot is started."

        aSymbol == #returnFromSnapshot ifTrue: [self restart].
        super update: aSymbol.
        ^self
! !

!RDoItServer class methodsFor: 'host management'!

allowHost:aHostname
        "allow a host"

        OkHosts add: aHostname.
        BadHosts remove:aHostname ifAbsent:[].
!

denyHost:aHostname
        "disallow a host"

        BadHosts add: aHostname.
        OkHosts remove:aHostname ifAbsent:[].
! !

!RDoItServer methodsFor: 'host management'!

confirmNewHost: hostName 
        "A previously unauthorized host is trying to connect."
        "Ask the user if we can authorize this new host."

        | answer |

        answer := DialogView confirm: ('RDOIT ok from host ' , 
                                       hostName printString , 
                                       '?') 
                             "initialAnswer: true".
        answer ifTrue: [OkHosts add: hostName]
               ifFalse: [BadHosts add: hostName].
        ^answer
!

showHosts: aStream 
        "Output the current hosts, normally to stdout."

        aStream crtab; show: 'Authorized Hosts '.
        aStream print: OkHosts.
        aStream crtab; show: 'Unauthorized Hosts '.
        aStream print: BadHosts.
        aStream cr; show: ''
! !

!RDoItServer methodsFor: 'handle client'!

handleClient: aSocket 
        "This method is forked off when a new connection arrives."

        | who hostName deniedBlock |

"claus:original
        who := aSocket getPeer.
        hostName := who hostName.
"
        hostName := aSocket getPeerName.

        deniedBlock := ['Denied rdoit attempt from: ' print . hostName printNL].
        (BadHosts includes: hostName)
                ifTrue: deniedBlock
                ifFalse: [(OkHosts includes: hostName)
                                ifTrue: [self handleOkClient: aSocket]
                                ifFalse: [(self confirmNewHost:hostName)
                                                ifTrue: [self handleOkClient: aSocket]
                                                ifFalse: deniedBlock]].
        aSocket close.
        ^nil
!

handleOkClient: aSocket 
        "We read through cr, and then doit."
        "Note: evaluate:logged: requires textOrString if logged = true."

        | exConn nStrm aCollection value|

"
        exConn := ExternalConnection ioAccessor: aSocket.
        nStrm := exConn readAppendStream.
        aCollection := nStrm through: Character cr.
"

"/        aCollection := aSocket through: Character cr.
        aCollection := aSocket nextLine.

"
'got request <' print. aCollection print. '>' printNL.
"
        "claus:
         I catch all errors - and send some diagnostics to stderr instead
        "
        SignalSet anySignal handle:[:ex |
            'an error occured during rdoit evaluation' errorPrintNL.
            ex return
        ] do:[
            value := Compiler evaluate: aCollection readStream 
        ].
        ^ value
!

loopForever
        "Called when snapshot is started, and at instance creation."
        "I create and listen to my socket."
        "Note: Someone else has already forked me, so I can block."
        "SEE ALSO: UnixSocketAccessor class howToImplementAServer."

        | newskt thisSocket |

        'RDoItServer started with:' printNL.
        self showHosts:Stdout.

        thisSocket := UnixSocketAccessor newTCPserverAtPort: self portNumber.
        thisSocket notNil ifTrue: [
            socket := thisSocket
        ] ifFalse: [
            'socket creation failed (port:' errorPrint. 
            self portNumber errorPrint. ')' errorPrintNL.
            ^ nil
        ].
        thisSocket listenFor: 5.

        newskt := true.
        [socket == thisSocket and: [newskt notNil]] whileTrue:[
            newskt := socket accept.
            newskt notNil ifTrue: [
"
'got a connection ...' printNL.
"
                ([self handleClient: newskt] forkAt: self handlePriority) name:'rdoit executor'.
            ]
        ].
        thisSocket notNil ifTrue: [thisSocket close].
        self release.
        ^nil
! !

!RDoItServer methodsFor: 'constants'!

handlePriority
        "The priority a handler process should run at."

        ^Processor lowIOPriority
!

serverPriority
        "The priority the server process should run at."

        ^Processor lowIOPriority
! !

!RDoItServer methodsFor: 'private'!

portNumber
        "Return my port number."
        "Default st80 R4 == 8004 (totally arbitrary)."

        myPortNumber isNil ifTrue: [^8004].
        ^myPortNumber
! !

!RDoItServer class methodsFor: 'class initialization'!

initialize
        "Initialize class variables."
        "RDoItServer initialize."

        BadHosts := Set new.
        OkHosts := Set new
! !

!RDoItServer class methodsFor: 'instance creation'!

start
        "Start an rdoit server on the default port."
        "RDoItServer start."

        ^self new startServer
!

startAt: aPortNumber 
        "Start an rdoit server on a specified port."
        "RDoItServer startAt: 8004."
        "RDoItServer startAt: 9104."

        ^self new startServerAt: aPortNumber
! !

!RDoItServer class methodsFor: 'server destruction'!

killAll
        "Destroy all instances of the server. (overkill)"
        "RDoItServer killAll."

        RDoItServer allInstances do: [:a | a release].
"claus: ORIGINAL
        RDoItServer allInstances do: [:a | a become: String new].
"
        RDoItServer allInstances do: [:a | a becomeNil].
        ^nil
! !

RDoItServer initialize!
